"""Frgal decoder."""
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Type

from google.protobuf.message import Message
from google.protobuf.pyext._message import (MessageMapContainer, RepeatedCompositeContainer, RepeatedScalarContainer,
                                            ScalarMapContainer)
from google.protobuf.struct_pb2 import Struct
from google.protobuf.timestamp_pb2 import Timestamp
from pytz import UTC

from .exceptions import DecoderError, FrgalException, UnknownError


class GrpcDecoder(object):
    """
    Generic grpc backend decoder.

    @ivar decoders: Dictionary of protobuf types and callables to decode them.
    @ivar exception_decoders: Dictionary of exception names and related exception classes.

    Decoding description:
        * Decoding is done recursively.
        * Primitive types are not decoded.
        * Empty string is replaced by None.
        * Repeated fields are decoded as arrays.
        * Map fields are decoded as dictionaries.
        * Messages are decoded as dictionaries.
    """

    def __init__(self):
        """Set common message type decoders."""
        self.decoders = {}  # type: Dict[Any, Callable]
        self.exception_decoders = {}  # type: Dict[str, Type[FrgalException]]

        for typ in (bool, int, float, bytes, type(None)):
            self.set_decoder(typ, lambda x: x)  # pragma: no cover

        self.set_decoder(str, self._decode_str)

        # There are two different GRPC types for lists depending on their content
        self.set_decoder(RepeatedScalarContainer, self._decode_list)
        self.set_decoder(RepeatedCompositeContainer, self._decode_list)

        # There are two different GRPC types for maps depending on their content
        self.set_decoder(ScalarMapContainer, self._decode_dict)
        self.set_decoder(MessageMapContainer, self._decode_dict)

        self.set_decoder(Struct, self._decode_struct)
        self.set_decoder(Timestamp, self._decode_timestamp)
        self.set_decoder(Message, self._decode_message)

    def set_decoder(self, object_type: Any, decode_func: Callable) -> None:
        """
        Set custom decoder.

        Args:
            object_type: Backend object type.
            decode_func (Callable): Callable getting backend value and returning decoded one.
        """
        self.decoders[object_type] = decode_func

    def set_exception_decoder(self, name: str, exception: Type[FrgalException]) -> None:
        """
        Set custom exception.

        Args:
            name: Backend exception name.
            exception: Python exception to be returned.
        """
        self.exception_decoders[name] = exception

    def decode(self, value: Any) -> Any:
        """
        Decode backend value.

        If there is custom decoder for the value type or its ancestors, it's used.
        Otherwise, exception is raised.
        """
        for cls in type(value).mro():
            if cls in self.decoders:
                return self.decoders[cls](value)

        raise DecoderError('Cannot decode unknown type {}'.format(type(value)))

    def decode_exception(self, value: Message) -> FrgalException:
        """
        Return appropriate exception type.

        Args:
            value: GRPC Exception Message
        """
        name = value.WhichOneof('Reasons')
        if name in self.exception_decoders:
            return self.exception_decoders[name]()
        else:
            return UnknownError(name)

    def _decode_str(self, value: str) -> Optional[str]:
        """Decode string."""
        if not value:
            return None
        return value

    def _decode_list(self, value: Iterable) -> list:
        """Decode list."""
        return [self.decode(x) for x in value]

    def _decode_dict(self, value: Mapping) -> dict:
        """Decode dict."""
        # Keys are primitive types, so they don't need to be decoded.
        return {a: self.decode(b) for a, b in value.items()}

    def _decode_message(self, value: Message) -> Any:
        """
        Decode GRPC Message object.

        Message without any fields defined, such as Empty, is decoded to None.
        Message with only one field defined is decoded to the decoding that field.
        Otherwise, Message is decoded to dictionary.
          * Keys of this dictionary are all Message fields names, whether they're set or not, except for oneof fields.
          * Only explicitely set oneof fields are decoded.
          * Fields are recursively decoded.
        """
        possible_fields = value.DESCRIPTOR.fields
        if len(possible_fields) == 0:
            return None
        elif len(possible_fields) == 1:
            return self.decode(getattr(value, possible_fields[0].name))
        else:
            decoded = {}  # type: Dict[str, Any]
            for field in possible_fields:
                if field.containing_oneof and field.name != value.WhichOneof(field.containing_oneof.name):
                    # Skip undefined oneof fields.
                    continue
                decoded[field.name] = self.decode(getattr(value, field.name))
            return decoded

    def _decode_struct(self, value: Struct) -> Dict:
        """Decode google.protobuf.Struct to Dict."""
        return {k: self.decode(v) for k, v in value.items()}

    @staticmethod
    def _decode_timestamp(timestamp: Timestamp) -> Optional[datetime]:
        """Decode timestamp to aware datetime or None."""
        if not timestamp.seconds and not timestamp.nanos:
            return None
        return UTC.localize(timestamp.ToDatetime())
